@@ -36,10 +36,20 @@
 #include <libxfce4ui/libxfce4ui.h>
 #include <libnotify/notify.h>
 
+#include "xfce4-notifyd/xfce-notify-log.h"
 #include "xfce4-notifyd-config.ui.h"
 
 #define KNOWN_APPLICATIONS_PROP       "/applications/known_applications"
 #define MUTED_APPLICATIONS_PROP       "/applications/muted_applications"
+#define LOG_DISPLAY_LIMIT             100
+
+typedef struct
+{
+    GtkWidget *log_level;
+    GtkWidget *log_level_apps;
+    GtkWidget *log_level_apps_label;
+    GtkWidget *infobar_label;
+} NotificationLogWidgets;
 
 static void
 xfce_notifyd_config_show_notification_callback(NotifyNotification *notification,
@@ -281,9 +291,9 @@ xfce4_notifyd_mute_application (GtkListBox *known_applications_listbox,
         muted_applications = g_ptr_array_new ();
 
     application_box = GTK_WIDGET (g_list_nth_data (gtk_container_get_children (GTK_CONTAINER (selected_application_row)), 0));
-    application_label = GTK_WIDGET (g_list_nth_data (gtk_container_get_children (GTK_CONTAINER (application_box)), 0));
+    application_label = GTK_WIDGET (g_list_nth_data (gtk_container_get_children (GTK_CONTAINER (application_box)), 1));
     application_name = gtk_label_get_text (GTK_LABEL(application_label));
-    mute_switch = GTK_WIDGET (g_list_nth_data (gtk_container_get_children (GTK_CONTAINER (application_box)), 1));
+    mute_switch = GTK_WIDGET (g_list_nth_data (gtk_container_get_children (GTK_CONTAINER (application_box)), 2));
     muted = gtk_switch_get_active (GTK_SWITCH (mute_switch));
 
     val = g_new0 (GValue, 1);
@@ -309,8 +319,6 @@ xfce4_notifyd_mute_application (GtkListBox *known_applications_listbox,
         g_warning ("Could not add %s to the list of muted applications.", new_app_name);
 
     xfconf_array_free (muted_applications);
-    /* FIXME: why does this cause a segfault? */
-    //g_free (new_app_name);
 }
 
 static void
@@ -323,7 +331,7 @@ xfce4_notifyd_row_activated (GtkListBox *known_applications_listbox,
     gboolean muted;
 
     application_box = GTK_WIDGET (g_list_nth_data (gtk_container_get_children (GTK_CONTAINER (selected_application_row)), 0));
-    mute_switch = GTK_WIDGET (g_list_nth_data (gtk_container_get_children (GTK_CONTAINER (application_box)), 1));
+    mute_switch = GTK_WIDGET (g_list_nth_data (gtk_container_get_children (GTK_CONTAINER (application_box)), 2));
     muted = !gtk_switch_get_active (GTK_SWITCH (mute_switch));
     gtk_switch_set_active (GTK_SWITCH (mute_switch), muted);
 }
@@ -364,13 +372,14 @@ xfce4_notifyd_known_applications_changed (XfconfChannel *channel,
     GtkWidget *known_applications_listbox = user_data;
     GtkWidget *hbox;
     GtkWidget *label;
+    GtkWidget *icon;
     GtkWidget *mute_switch;
-    GtkWidget *separator;
     GtkCallback func = listbox_remove_all;
     GPtrArray *known_applications;
     GPtrArray *muted_applications;
     GValue *known_application;
     guint i, j;
+    const gchar *icon_name;
 
     known_applications = xfconf_channel_get_arrayv (channel, KNOWN_APPLICATIONS_PROP);
     muted_applications = xfconf_channel_get_arrayv (channel, MUTED_APPLICATIONS_PROP);
@@ -385,9 +394,25 @@ xfce4_notifyd_known_applications_changed (XfconfChannel *channel,
             known_application = g_ptr_array_index (known_applications, i);
             hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 6);
             label = gtk_label_new (g_value_get_string (known_application));
+            icon_name = g_value_get_string (known_application);
+            if (gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default(), icon_name, 16, GTK_ICON_LOOKUP_GENERIC_FALLBACK))
+                icon = gtk_image_new_from_icon_name (icon_name, GTK_ICON_SIZE_MENU);
+            else if (gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default(), g_ascii_strdown(icon_name, -1), 16, GTK_ICON_LOOKUP_GENERIC_FALLBACK))
+                icon = gtk_image_new_from_icon_name (g_ascii_strdown(icon_name, -1), GTK_ICON_SIZE_MENU);
+            else {
+                icon = gtk_image_new ();
+                gtk_image_set_pixel_size (GTK_IMAGE(icon), 16);
+            }
+
+#if GTK_CHECK_VERSION (3, 16, 0)
             gtk_label_set_xalign (GTK_LABEL (label), 0);
+#else
+            gtk_widget_set_halign (label, GTK_ALIGN_START);
+#endif
             mute_switch = gtk_switch_new ();
+            gtk_widget_set_valign (label, GTK_ALIGN_CENTER);
             gtk_switch_set_active (GTK_SWITCH (mute_switch), TRUE);
+            gtk_box_pack_start (GTK_BOX (hbox), icon, FALSE, TRUE, 3);
             gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 3);
             gtk_box_pack_end (GTK_BOX (hbox), mute_switch, FALSE, TRUE, 3);
             gtk_list_box_insert (GTK_LIST_BOX (known_applications_listbox), hbox, -1);
@@ -424,12 +449,182 @@ xfce4_notifyd_do_not_disturb_activated (GtkSwitch *do_not_disturb_switch,
     gtk_switch_set_active (GTK_SWITCH (do_not_disturb_switch), state);
 }
 
+
+static void
+xfce4_notifyd_log_activated (GtkSwitch *log_switch,
+                             gboolean state,
+                             gpointer user_data)
+{
+    NotificationLogWidgets *log_widgets = user_data;
+    const char *format = "<b>Currently only urgent notifications are shown.</b>\nNotification logging is \%s.";
+    char *markup;
+
+    gtk_switch_set_state (GTK_SWITCH (log_switch), state);
+    gtk_widget_set_sensitive (GTK_WIDGET (log_widgets->log_level), state);
+    gtk_widget_set_sensitive (GTK_WIDGET (log_widgets->log_level_apps), state);
+    gtk_widget_set_sensitive (GTK_WIDGET (log_widgets->log_level_apps_label), state);
+    markup = g_markup_printf_escaped (format, state ? "enabled" : "disabled");
+    gtk_label_set_markup (GTK_LABEL (log_widgets->infobar_label), markup);
+    g_free (markup);
+}
+
+static void
+xfce4_notifyd_log_open (GtkButton *button, gpointer user_data) {
+    gchar *notify_log_path = NULL;
+
+    notify_log_path = xfce_resource_lookup (XFCE_RESOURCE_CACHE, XFCE_NOTIFY_LOG_FILE);
+    if (!g_app_info_launch_default_for_uri (g_strdup_printf("file://%s", notify_log_path), NULL, NULL))
+        g_warning ("Could not open the log file: %s", notify_log_path);
+}
+
+static void
+xfce4_notifyd_log_populate (GtkWidget *log_listbox)
+{
+    GKeyFile *notify_log;
+    GtkCallback func = listbox_remove_all;
+    gint i;
+    GDateTime *today;
+    gchar *timestamp;
+    gchar *limit_label;
+    GtkWidget *limit_button;
+
+    today = g_date_time_new_now_local ();
+    timestamp = g_date_time_format (today, "%F");
+
+    gtk_container_foreach (GTK_CONTAINER (log_listbox), func, log_listbox);
+    notify_log = xfce_notify_log_get();
+
+    if (notify_log) {
+        gchar **groups;
+        gboolean yesterday = FALSE;
+        gsize num_groups;
+        int log_length;
+
+        groups = g_key_file_get_groups (notify_log, &num_groups);
+        log_length = GPOINTER_TO_UINT(num_groups) - LOG_DISPLAY_LIMIT;
+        if (log_length < 0)
+            log_length = 0;
+
+        /* Notifications are only shown until LOG_DISPLAY_LIMIT is hit */
+        for (i = log_length; groups && groups[i]; i += 1) {
+            GtkWidget *grid;
+            GtkWidget *summary, *body, *app_icon, *expire_timeout;
+            const gchar *group = groups[i];
+            const char *format = "<b>\%s</b>";
+            const char *tooltip_format = "<b>\%s</b> - \%s\n\%s";
+            const char *tooltip_format_simple = "<b>\%s</b> - \%s";
+            char *markup;
+            gchar *app_name;
+            gchar *tooltip_timestamp;
+            GTimeVal tv;
+            GDateTime *log_timestamp;
+
+            if (g_time_val_from_iso8601 (group, &tv) == TRUE) {
+                if (log_timestamp = g_date_time_new_from_timeval_local (&tv)) {
+                    tooltip_timestamp = g_date_time_format (log_timestamp, "%c");
+                    g_date_time_unref(log_timestamp);
+                }
+            }
+
+            if (g_ascii_strncasecmp (timestamp, group, 10) == 0 && yesterday == FALSE) {
+                GtkWidget *header;
+                header = gtk_label_new ("Yesterday and before");
+                gtk_widget_set_sensitive (header, FALSE);
+                gtk_widget_set_margin_top (header, 3);
+                gtk_widget_set_margin_bottom (header, 3);
+                gtk_list_box_insert (GTK_LIST_BOX (log_listbox), header, 0);
+                yesterday = TRUE;
+            }
+
+            app_name = g_key_file_get_string (notify_log, group, "app_name", NULL);
+            markup = g_markup_printf_escaped (format, g_key_file_get_string (notify_log, group, "summary", NULL));
+            summary = gtk_label_new (NULL);
+            gtk_label_set_markup (GTK_LABEL (summary), markup);
+            gtk_label_set_xalign (GTK_LABEL (summary), 0);
+            g_free (markup);
+            body = gtk_label_new (g_key_file_get_string (notify_log, group, "body", NULL));
+            gtk_label_set_xalign (GTK_LABEL (body), 0);
+            gtk_label_set_ellipsize (GTK_LABEL (body), PANGO_ELLIPSIZE_END);
+            app_icon = gtk_image_new_from_icon_name (g_key_file_get_string (notify_log, group, "app_icon", NULL),
+                                                     GTK_ICON_SIZE_LARGE_TOOLBAR);
+            gtk_image_set_pixel_size (GTK_IMAGE (app_icon), 24);
+            gtk_widget_set_margin_start (app_icon, 3);
+            expire_timeout = gtk_label_new (g_key_file_get_string (notify_log, group, "expire-timeout", NULL));
+            // TODO: actions and timeout are missing (timeout is only interesting for urgent messages) - do we need that?
+            grid = gtk_grid_new ();
+            gtk_grid_set_column_spacing (GTK_GRID (grid), 6);
+            gtk_grid_attach (GTK_GRID (grid), GTK_WIDGET (app_icon), 0, 0 , 1, 2);
+
+            /* Handle icon-only notifications */
+            if (g_strcmp0 (g_key_file_get_string (notify_log, group, "body", NULL), "") == 0) {
+                gtk_grid_attach (GTK_GRID (grid), GTK_WIDGET (summary), 1, 0, 1, 2);
+                markup = g_strdup_printf (tooltip_format_simple, app_name, tooltip_timestamp);
+            }
+            else {
+                gtk_grid_attach (GTK_GRID (grid), GTK_WIDGET (summary), 1, 0, 1, 1);
+                gtk_grid_attach (GTK_GRID (grid), GTK_WIDGET (body), 1, 1, 1, 1);
+                markup = g_strdup_printf (tooltip_format, app_name, tooltip_timestamp,
+                                          g_key_file_get_string (notify_log, group, "body", NULL));
+            }
+
+            gtk_widget_set_tooltip_markup (grid, markup);
+            g_free (markup);
+
+            gtk_list_box_insert (GTK_LIST_BOX (log_listbox), grid, 0);
+        }
+        if (log_length > 0) {
+            limit_label = g_strdup_printf("Showing %d of %d notifications. Click to open full log.",
+                                          LOG_DISPLAY_LIMIT, GPOINTER_TO_UINT(num_groups));
+            limit_button = gtk_button_new_with_label (limit_label);
+            gtk_widget_set_margin_bottom (limit_button, 3);
+            gtk_widget_set_margin_top (limit_button, 3);
+            gtk_button_set_relief (GTK_BUTTON (limit_button), GTK_RELIEF_NONE);
+            g_signal_connect (G_OBJECT (limit_button), "clicked",
+                              G_CALLBACK (xfce4_notifyd_log_open), log_listbox);
+            gtk_list_box_insert (GTK_LIST_BOX (log_listbox), limit_button, LOG_DISPLAY_LIMIT + 1);
+        }
+        g_key_file_free (notify_log);
+    }
+
+    gtk_widget_show_all (log_listbox);
+}
+
+static void
+xfce4_notifyd_log_refresh (GtkButton *button, gpointer user_data) {
+    GtkWidget *log_listbox = user_data;
+    xfce4_notifyd_log_populate (log_listbox);
+}
+
+static void xfce_notify_log_clear_button_clicked (GtkButton *button, gpointer user_data) {
+    GtkWidget *log_listbox = user_data;
+    GtkCallback func = listbox_remove_all;
+
+    /* Clear the log file and empty the listbox widget */
+    gtk_container_foreach (GTK_CONTAINER (log_listbox), func, log_listbox);
+    xfce_notify_log_clear();
+}
+
 static void xfce4_notifyd_show_help(GtkButton *button,
                                     GtkWidget *dialog)
 {
     xfce_dialog_show_help_with_version(GTK_WINDOW(dialog), "notifyd", "start", NULL, NULL);
 }
 
+GtkWidget *
+placeholder_label_new (gchar *place_holder_text)
+{
+    GtkWidget *label;
+    label = gtk_label_new ("");
+    gtk_label_set_markup (GTK_LABEL (label), place_holder_text);
+    gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_CENTER);
+    gtk_widget_set_sensitive (label, FALSE);
+    gtk_widget_set_margin_start (label, 24);
+    gtk_widget_set_margin_end (label, 24);
+    gtk_widget_set_margin_top (label, 24);
+    gtk_widget_set_margin_bottom (label, 24);
+    return label;
+}
+
 static GtkWidget *
 xfce4_notifyd_config_setup_dialog(GtkBuilder *builder)
 {
@@ -446,9 +641,21 @@ xfce4_notifyd_config_setup_dialog(GtkBuilder *builder)
     GtkWidget *placeholder_label;
     GtkWidget *do_not_disturb_switch;
     GtkWidget *do_not_disturb_info;
+    GtkWidget *log_switch;
+    GtkWidget *log_scrolled_window;
+    GtkWidget *log_listbox;
+    GtkToolbar *log_toolbar;
+    GtkToolItem *log_clear_button;
+    GtkToolItem *log_refresh_button;
+    GtkToolItem *log_open_button;
+    GtkWidget *icon;
+    GtkWidget *primary_monitor;
+    GtkWidget *do_fadeout;
     GtkAdjustment *adj;
     GError *error = NULL;
     gchar *current_theme;
+    GKeyFile *notification_log;
+    NotificationLogWidgets log_widgets;
 
     gtk_builder_connect_signals(builder, NULL);
 
@@ -460,9 +667,9 @@ xfce4_notifyd_config_setup_dialog(GtkBuilder *builder)
     g_signal_connect_swapped(G_OBJECT(btn), "clicked",
                              G_CALLBACK(gtk_dialog_response), dlg);
 
-    btn = GTK_WIDGET(gtk_builder_get_object(builder, "preview_button"));
-    g_signal_connect(G_OBJECT(btn), "clicked",
-                     G_CALLBACK(xfce_notifyd_config_preview_clicked), dlg);
+    help_button = GTK_WIDGET(gtk_builder_get_object(builder, "help_btn"));
+    g_signal_connect(G_OBJECT(help_button), "clicked",
+                    G_CALLBACK(xfce4_notifyd_show_help), dlg);
 
     if(!xfconf_init(&error)) {
         xfce_message_dialog(NULL, _("Xfce Notify Daemon"),
@@ -476,22 +683,29 @@ xfce4_notifyd_config_setup_dialog(GtkBuilder *builder)
 
     channel = xfconf_channel_new("xfce4-notifyd");
 
-    sbtn = GTK_WIDGET(gtk_builder_get_object(builder, "expire_timeout_sbtn"));
-    xfconf_g_property_bind(channel, "/expire-timeout", G_TYPE_INT,
-                           G_OBJECT(sbtn), "value");
+    /**************
+        GENERAL   *
+     **************/
+    // Behavior
+    do_not_disturb_switch = GTK_WIDGET (gtk_builder_get_object (builder, "do_not_disturb"));
+    xfconf_g_property_bind (channel, "/do-not-disturb", G_TYPE_BOOLEAN,
+                            G_OBJECT (do_not_disturb_switch), "active");
+    /* Manually control the revealer for the infobar because of https://bugzilla.gnome.org/show_bug.cgi?id=710888 */
+    do_not_disturb_info = GTK_WIDGET (gtk_builder_get_object (builder, "do_not_disturb_info"));
+    gtk_revealer_set_reveal_child (GTK_REVEALER (do_not_disturb_info),
+                                   gtk_switch_get_active (GTK_SWITCH (do_not_disturb_switch)));
+    g_signal_connect (G_OBJECT (do_not_disturb_switch), "state-set",
+                      G_CALLBACK (xfce4_notifyd_do_not_disturb_activated), do_not_disturb_info);
 
-    slider = GTK_WIDGET(gtk_builder_get_object(builder, "opacity_slider"));
-    g_signal_connect(G_OBJECT(slider), "format-value",
-                     G_CALLBACK(xfce4_notifyd_slider_format_value), NULL);
-    adj = gtk_range_get_adjustment(GTK_RANGE(slider));
-    xfconf_g_property_bind(channel, "/initial-opacity", G_TYPE_DOUBLE,
-                           G_OBJECT(adj), "value");
+    primary_monitor = GTK_WIDGET(gtk_builder_get_object(builder, "primary_monitor"));
+    xfconf_g_property_bind(channel, "/primary-monitor", G_TYPE_UINT,
+                           G_OBJECT(primary_monitor), "active");
 
+    // Appearance
     theme_combo = GTK_WIDGET(gtk_builder_get_object(builder, "theme_combo"));
     current_theme = xfconf_channel_get_string(channel, "/theme", "Default");
     xfce4_notifyd_config_setup_theme_combo(theme_combo, current_theme);
     g_free(current_theme);
-
     g_signal_connect(G_OBJECT(theme_combo), "changed",
                      G_CALLBACK(xfce4_notifyd_config_theme_combo_changed),
                      channel);
@@ -505,30 +719,36 @@ xfce4_notifyd_config_setup_dialog(GtkBuilder *builder)
     if(gtk_combo_box_get_active(GTK_COMBO_BOX(position_combo)) == -1)
         gtk_combo_box_set_active(GTK_COMBO_BOX(position_combo), GTK_CORNER_TOP_RIGHT);
 
-    do_not_disturb_switch = GTK_WIDGET (gtk_builder_get_object (builder, "do_not_disturb"));
-    xfconf_g_property_bind (channel, "/do-not-disturb", G_TYPE_BOOLEAN,
-                            G_OBJECT (do_not_disturb_switch), "active");
-    /* Manually control the revealer for the infobar because of https://bugzilla.gnome.org/show_bug.cgi?id=710888 */
-    do_not_disturb_info = GTK_WIDGET (gtk_builder_get_object (builder, "do_not_disturb_info"));
-    gtk_revealer_set_reveal_child (GTK_REVEALER (do_not_disturb_info),
-                                   gtk_switch_get_active (GTK_SWITCH (do_not_disturb_switch)));
-    g_signal_connect (G_OBJECT (do_not_disturb_switch), "state-set",
-                      G_CALLBACK (xfce4_notifyd_do_not_disturb_activated), do_not_disturb_info);
+    slider = GTK_WIDGET(gtk_builder_get_object(builder, "opacity_slider"));
+    g_signal_connect(G_OBJECT(slider), "format-value",
+                     G_CALLBACK(xfce4_notifyd_slider_format_value), NULL);
+    adj = gtk_range_get_adjustment(GTK_RANGE(slider));
+    xfconf_g_property_bind(channel, "/initial-opacity", G_TYPE_DOUBLE,
+                           G_OBJECT(adj), "value");
 
+    sbtn = GTK_WIDGET(gtk_builder_get_object(builder, "expire_timeout_sbtn"));
+    xfconf_g_property_bind(channel, "/expire-timeout", G_TYPE_INT,
+                           G_OBJECT(sbtn), "value");
+
+    do_fadeout = GTK_WIDGET(gtk_builder_get_object(builder, "do_fadeout"));
+    xfconf_g_property_bind(channel, "/do-fadeout", G_TYPE_BOOLEAN,
+                          G_OBJECT(do_fadeout), "active");
+
+    btn = GTK_WIDGET(gtk_builder_get_object(builder, "preview_button"));
+    g_signal_connect(G_OBJECT(btn), "clicked",
+                     G_CALLBACK(xfce_notifyd_config_preview_clicked), dlg);
+
+    /*******************
+        APPLICATIONS   *
+     *******************/
     known_applications_scrolled_window = GTK_WIDGET (gtk_builder_get_object (builder, "known_applications_scrolled_window"));
     known_applications_listbox = gtk_list_box_new ();
     gtk_container_add (GTK_CONTAINER (known_applications_scrolled_window), known_applications_listbox);
     gtk_list_box_set_header_func (GTK_LIST_BOX (known_applications_listbox), display_header_func, NULL, NULL);
 
-    placeholder_label = gtk_label_new ("");
-    gtk_label_set_markup (GTK_LABEL (placeholder_label),"<b>Currently there are no known applications.</b>\nAs soon as an application sends a notification\nit will appear in this list.");
-    gtk_label_set_justify (GTK_LABEL (placeholder_label), GTK_JUSTIFY_CENTER);
-    gtk_widget_set_sensitive (placeholder_label, FALSE);
-    gtk_widget_set_margin_start (placeholder_label, 24);
-    gtk_widget_set_margin_end (placeholder_label, 24);
-    gtk_widget_set_margin_top (placeholder_label, 24);
-    gtk_widget_set_margin_bottom (placeholder_label, 24);
-
+    placeholder_label = placeholder_label_new ("<big><b>Currently there are no known applications.</b></big>"
+                                                "\nAs soon as an application sends a notification"
+                                                "\nit will appear in this list.");
     /* Initialize the list of known applications */
     xfce4_notifyd_known_applications_changed (channel, KNOWN_APPLICATIONS_PROP, NULL, known_applications_listbox);
     gtk_list_box_set_placeholder (GTK_LIST_BOX (known_applications_listbox), placeholder_label);
@@ -540,9 +760,60 @@ xfce4_notifyd_config_setup_dialog(GtkBuilder *builder)
                       "property-changed::" KNOWN_APPLICATIONS_PROP,
                       G_CALLBACK (xfce4_notifyd_known_applications_changed), known_applications_listbox);
 
-    help_button = GTK_WIDGET(gtk_builder_get_object(builder, "help_btn"));
-    g_signal_connect(G_OBJECT(help_button), "clicked",
-                     G_CALLBACK(xfce4_notifyd_show_help), dlg);
+    /**********
+        LOG   *
+     **********/
+    log_widgets.log_level = GTK_WIDGET (gtk_builder_get_object (builder, "log_level"));
+    log_widgets.log_level_apps = GTK_WIDGET (gtk_builder_get_object (builder, "log_level_apps"));
+    log_widgets.log_level_apps_label = GTK_WIDGET (gtk_builder_get_object (builder, "log_level_apps_label"));
+    log_widgets.infobar_label = GTK_WIDGET (gtk_builder_get_object (builder, "infobar_label"));
+    log_switch = GTK_WIDGET (gtk_builder_get_object (builder, "log_switch"));
+    xfconf_g_property_bind (channel, "/notification-log", G_TYPE_BOOLEAN,
+                            G_OBJECT (log_switch), "active");
+    g_signal_connect (G_OBJECT (log_switch), "state-set",
+                      G_CALLBACK (xfce4_notifyd_log_activated), &log_widgets);
+    xfconf_g_property_bind(channel, "/log-level", G_TYPE_UINT,
+                           G_OBJECT(log_widgets.log_level), "active");
+    xfconf_g_property_bind(channel, "/log-level-apps", G_TYPE_UINT,
+                          G_OBJECT(log_widgets.log_level_apps), "active");
+
+    /* Initialize the settings' states correctly */
+    if(gtk_combo_box_get_active(GTK_COMBO_BOX(log_widgets.log_level)) == -1)
+        gtk_combo_box_set_active(GTK_COMBO_BOX(log_widgets.log_level), 0);
+    if(gtk_combo_box_get_active(GTK_COMBO_BOX(log_widgets.log_level_apps)) == -1)
+        gtk_combo_box_set_active(GTK_COMBO_BOX(log_widgets.log_level_apps), 0);
+    xfce4_notifyd_log_activated (GTK_SWITCH (log_switch), gtk_switch_get_active (GTK_SWITCH(log_switch)), &log_widgets);
+
+    log_scrolled_window = GTK_WIDGET (gtk_builder_get_object (builder, "log_scrolled_window"));
+    log_listbox = gtk_list_box_new ();
+    gtk_container_add (GTK_CONTAINER (log_scrolled_window), log_listbox);
+    gtk_list_box_set_header_func (GTK_LIST_BOX (log_listbox), display_header_func, NULL, NULL);
+    placeholder_label = placeholder_label_new ("<big><b>Empty log</b></big>"
+                                               "\nNo notifications have been logged yet.");
+    gtk_list_box_set_placeholder (GTK_LIST_BOX (log_listbox), placeholder_label);
+    gtk_widget_show_all (placeholder_label);
+    xfce4_notifyd_log_populate (log_listbox);
+
+    log_toolbar = GTK_TOOLBAR (gtk_builder_get_object (builder, "log_toolbar"));
+    icon = gtk_image_new_from_icon_name ("view-refresh-symbolic", GTK_ICON_SIZE_LARGE_TOOLBAR);
+    log_refresh_button = gtk_tool_button_new (icon, _("Refresh"));
+    gtk_widget_set_tooltip_text (GTK_WIDGET (log_refresh_button), _("Refresh the notification log"));
+    gtk_toolbar_insert(log_toolbar, GTK_TOOL_ITEM(log_refresh_button), 0);
+    g_signal_connect (G_OBJECT (log_refresh_button), "clicked",
+                      G_CALLBACK (xfce4_notifyd_log_refresh), log_listbox);
+    icon = gtk_image_new_from_icon_name ("document-open-symbolic", GTK_ICON_SIZE_LARGE_TOOLBAR);
+    log_open_button = gtk_tool_button_new (icon, _("Open"));
+    gtk_widget_set_tooltip_text (GTK_WIDGET (log_open_button), _("Open the notification log in an external editor"));
+    gtk_toolbar_insert(log_toolbar, GTK_TOOL_ITEM(log_open_button), 1);
+    g_signal_connect (G_OBJECT (log_open_button), "clicked",
+                      G_CALLBACK (xfce4_notifyd_log_open), log_listbox);
+    icon = gtk_image_new_from_icon_name ("edit-clear-symbolic", GTK_ICON_SIZE_LARGE_TOOLBAR);
+    log_clear_button = gtk_tool_button_new (icon, _("Clear"));
+    gtk_widget_set_tooltip_text (GTK_WIDGET (log_clear_button), _("Clear the notification log"));
+    gtk_toolbar_insert(log_toolbar, GTK_TOOL_ITEM(log_clear_button), 2);
+    g_signal_connect (G_OBJECT (log_clear_button), "clicked",
+                      G_CALLBACK (xfce_notify_log_clear_button_clicked), log_listbox);
+    gtk_widget_show_all (GTK_WIDGET(log_toolbar));
 
     return dlg;
 }
@@ -615,7 +886,6 @@ main(int argc,
         /* In the glade file, plug_child has parent, so remove it first */
         gtk_container_remove (GTK_CONTAINER(gtk_widget_get_parent(plug_child)), plug_child);
         gtk_container_add (GTK_CONTAINER(plug), plug_child);
-        gtk_widget_set_valign (plug_child, GTK_ALIGN_START);
 
         gdk_notify_startup_complete();
         g_object_unref(G_OBJECT(builder));
